from sumo.wrapper import SumoClient
from fmu.sumo.explorer.objects.surface_collection import SurfaceCollection
from fmu.sumo.explorer.objects.polygons_collection import PolygonsCollection
from fmu.sumo.explorer.objects.table_collection import TableCollection


REALIZATION_FILTER = {
    "bool": {
        "must": [
            {"exists": {"field": "fmu.iteration.name.keyword"}},
            {"exists": {"field": "fmu.realization.id"}},
        ]
    }
}


class RealizationContext:
    def __init__(self, sumo: SumoClient, case_id: str) -> None:
        self._sumo = sumo
        self._case_id = case_id

    @property
    def surfaces(self) -> SurfaceCollection:
        return SurfaceCollection(self._sumo, self._case_id, REALIZATION_FILTER)

    @property
    def polygons(self) -> PolygonsCollection:
        return PolygonsCollection(self._sumo, self._case_id, REALIZATION_FILTER)

    @property
    def tables(self) -> TableCollection:
        return TableCollection(self._sumo, self._case_id, REALIZATION_FILTER)



"""Contains class for one document"""
from typing import List, Dict


class Document:
    """Class for representing a document in Sumo"""

    def __init__(self, metadata: Dict) -> None:
        self._uuid = metadata["_id"]
        self._metadata = metadata["_source"]

    @property
    def uuid(self):
        """Return uuid

        Returns:
            str: the uuid of the case
        """
        return self._uuid

    @property
    def metadata(self):
        """Return metadata for document

        Returns:
            dict: the metadata
        """
        return self._metadata

    def _get_property(self, path: List[str]):
        curr = self._metadata.copy()

        for key in path:
            if key in curr:
                curr = curr[key]
            else:
                return None

        return curr

    def __getitem__(self, key: str):
        return self._metadata[key]



"""Module containing class for pit handling"""
from typing import Dict
from sumo.wrapper import SumoClient


class Pit:
    """Class for handling of pit"""

    def __init__(self, sumo: SumoClient, keep_alive: str) -> None:
        """Init

        Args:
            sumo (SumoClient): Activated sumo client
            keep_alive (str): how long to keep instance alive
        """
        self._sumo = sumo
        self._keep_alive = keep_alive
        self._pit_id = self.__get_pit_id(keep_alive)

    def __get_pit_id(self, keep_alive) -> str:
        res = self._sumo.post("/pit", params={"keep-alive": keep_alive})
        return res.json()["id"]

    def get_pit_object(self, pit_id: str = None) -> Dict:
        """Get the pit object

        Returns:
            Dict: dict with id and info about how long to keep alive
        """
        return {
            "id": pit_id if pit_id is not None else self._pit_id,
            "keep_alive": self._keep_alive,
        }



"""Module containing class for exploring results from sumo"""
from sumo.wrapper import SumoClient
from fmu.sumo.explorer.pit import Pit
from fmu.sumo.explorer.objects.case_collection import (
    CaseCollection,
    _CASE_FIELDS,
)
from fmu.sumo.explorer.objects._child_collection import _CHILD_FIELDS
from fmu.sumo.explorer.objects.surface import Surface
from fmu.sumo.explorer.objects.polygons import Polygons
from fmu.sumo.explorer.objects.table import Table
from fmu.sumo.explorer.objects.case import Case
from fmu.sumo.explorer._utils import Utils


class Explorer:
    """Class for consuming FMU results from Sumo.
    The Sumo Explorer is a Python package for consuming FMU results stored
    in Sumo. It is FMU aware, and creates an abstraction on top of the
    Sumo API. The purpose of the package is to create an FMU-oriented
    Python interface towards FMU data in Sumo, and make it easy for FMU
    users in various contexts to use data stored in Sumo.

    Examples of use cases:
      - Applications (example: Webviz)
      - Scripts (example: Local post-processing functions)
      - Manual data browsing and visualization (example: A Jupyter Notebook)
    """

    def __init__(
        self,
        env: str = "prod",
        token: str = None,
        interactive: bool = True,
        keep_alive: str = None,
    ):
        """Initialize the Explorer class

        When iterating over large datasets, use the `keep_alive` argument
        to create a snapshot of the data to ensure consistency. The
        argument specifies the lifespan of the snapshot and every
        request to the Sumo API will extend the lifetime of the snapshot
        with the specified `keep_alive` value. The argument uses a format
        of a number followed by a unit indicator. Supported indicators are:
            - d (day)
            - h (hour)
            - m (minute)
            - s (second)
            - ms (milisecond)
            - micros (microsecond)
            - nanos (nanosecond)

        Examples: 1d, 2h, 15m, 30s

        Every request to Sumo will extend the lifespan of the snapshot
        by the time specified in `keep_alive`.

        Args:
            env (str): Sumo environment
            token (str): authenticate with existing token
            interactive (bool): authenticate using interactive flow (browser)
            keep_alive (str): point in time lifespan
        """
        self._sumo = SumoClient(env, token=token, interactive=interactive)
        self._pit = Pit(self._sumo, keep_alive) if keep_alive else None
        self._utils = Utils(self._sumo)

    @property
    def cases(self):
        """Cases in Sumo"""
        return CaseCollection(sumo=self._sumo, pit=self._pit)

    def get_permissions(self, asset: str = None):
        """Get permissions

        Args:
            asset (str): asset in Sumo

        Returns:
          dict: Dictionary of user permissions
        """
        res = self._sumo.get("/userpermissions").json()

        if asset is not None:
            if asset not in res:
                raise PermissionError(f"No permissions for asset: {asset}")

        return res

    async def get_permissions_async(self, asset: str = None):
        """Get permissions

        Args:
            asset (str): asset in Sumo

        Returns:
          dict: Dictionary of user permissions
        """
        res = await self._sumo.get_async("/userpermissions")
        res = res.json()

        if asset is not None:
            if asset not in res:
                raise PermissionError(f"No permissions for asset: {asset}")

        return res

    def get_case_by_uuid(self, uuid: str) -> Case:
        """Get case object by uuid

        Args:
            uuid (str): case uuid

        Returns:
            Case: case object
        """
        cases = self.cases.filter(uuid=uuid)
        if len(cases) == 0:
            raise Exception(f"Document not found: {uuid}")

        return cases[0]

    async def get_case_by_uuid_async(self, uuid: str) -> Case:
        """Get case object by uuid

        Args:
            uuid (str): case uuid

        Returns:
            Case: case object
        """
        cases = self.cases.filter(uuid=uuid)
        if await cases.length_async() == 0:
            raise Exception(f"Document not found: {uuid}")

        return await cases.getitem_async(0)

    def get_surface_by_uuid(self, uuid: str) -> Surface:
        """Get surface object by uuid

        Args:
            uuid (str): surface uuid

        Returns:
            Surface: surface object
        """
        metadata = self._utils.get_object(uuid, _CHILD_FIELDS)
        return Surface(self._sumo, metadata)

    async def get_surface_by_uuid_async(self, uuid: str) -> Surface:
        """Get surface object by uuid

        Args:
            uuid (str): surface uuid

        Returns:
            Surface: surface object
        """
        metadata = await self._utils.get_object_async(uuid, _CHILD_FIELDS)
        return Surface(self._sumo, metadata)

    def get_polygons_by_uuid(self, uuid: str) -> Polygons:
        """Get polygons object by uuid

        Args:
            uuid (str): polygons uuid

        Returns:
            Polygons: polygons object
        """
        metadata = self._utils.get_object(uuid, _CHILD_FIELDS)
        return Polygons(self._sumo, metadata)

    async def get_polygons_by_uuid_async(self, uuid: str) -> Polygons:
        """Get polygons object by uuid

        Args:
            uuid (str): polygons uuid

        Returns:
            Polygons: polygons object
        """
        metadata = await self._utils.get_object_async(uuid, _CHILD_FIELDS)
        return Polygons(self._sumo, metadata)

    def get_table_by_uuid(self, uuid: str) -> Table:
        """Get table object by uuid

        Args:
            uuid (str): table uuid

        Returns:
            Table: table object
        """
        metadata = self._utils.get_object(uuid, _CHILD_FIELDS)
        return Table(self._sumo, metadata)

    async def get_table_by_uuid_async(self, uuid: str) -> Table:
        """Get table object by uuid

        Args:
            uuid (str): table uuid

        Returns:
            Table: table object
        """
        metadata = await self._utils.get_object_async(uuid, _CHILD_FIELDS)
        return Table(self._sumo, metadata)



"""Module containing class for collection of children"""

from typing import List, Dict, Union
from sumo.wrapper import SumoClient
from fmu.sumo.explorer.objects._document_collection import DocumentCollection
from fmu.sumo.explorer.timefilter import TimeFilter
from fmu.sumo.explorer.pit import Pit

_CHILD_FIELDS = {
    "include": [],
    "exclude": ["data.spec.columns", "fmu.realization.parameters"],
}


class ChildCollection(DocumentCollection):
    """Class for representing a collection of child objects in Sumo"""

    def __init__(
        self,
        doc_type: str,
        sumo: SumoClient,
        case_uuid: str,
        query: Dict = None,
        pit: Pit = None,
    ):
        self._case_uuid = case_uuid
        super().__init__(doc_type, sumo, query, _CHILD_FIELDS, pit)

    @property
    def names(self) -> List[str]:
        """List of unique object names"""
        return self._get_field_values("data.name.keyword")

    @property
    async def names_async(self) -> List[str]:
        """List of unique object names"""
        return await self._get_field_values_async("data.name.keyword")

    @property
    def tagnames(self) -> List[str]:
        """List of unqiue object tagnames"""
        return self._get_field_values("data.tagname.keyword")

    @property
    async def tagnames_async(self) -> List[str]:
        """List of unqiue object tagnames"""
        return await self._get_field_values_async("data.tagname.keyword")

    @property
    def dataformats(self) -> List[str]:
        """List of unique data.format values"""
        return self._get_field_values("data.format.keyword")

    @property
    async def dataformats_async(self) -> List[str]:
        """List of unique data.format values"""
        return await self._get_field_values_async("data.format.keyword")

    @property
    def iterations(self) -> List[int]:
        """List of unique object iteration names"""
        return self._get_field_values("fmu.iteration.name.keyword")

    @property
    async def iterations_async(self) -> List[int]:
        """List of unique object iteration names"""
        return await self._get_field_values_async("fmu.iteration.name.keyword")

    @property
    def realizations(self) -> List[int]:
        """List of unique object realization ids"""
        return self._get_field_values("fmu.realization.id")

    @property
    async def realizations_async(self) -> List[int]:
        """List of unique object realization ids"""
        return await self._get_field_values_async("fmu.realization.id")

    @property
    def aggregations(self) -> List[str]:
        """List of unique object aggregation operations"""
        return self._get_field_values("fmu.aggregation.operation.keyword")

    @property
    async def aggregations_async(self) -> List[str]:
        """List of unique object aggregation operations"""
        return await self._get_field_values_async(
            "fmu.aggregation.operation.keyword"
        )

    @property
    def stages(self) -> List[str]:
        """List of unique stages"""
        return self._get_field_values("fmu.context.stage.keyword")

    @property
    async def stages_async(self) -> List[str]:
        """List of unique stages"""
        return await self._get_field_values_async("fmu.context.stage.keyword")

    @property
    def stratigraphic(self) -> List[str]:
        """List of unqiue object stratigraphic"""
        return self._get_field_values("data.stratigraphic")

    @property
    async def stratigraphic_async(self) -> List[str]:
        """List of unqiue object stratigraphic"""
        return await self._get_field_values_async("data.stratigraphic")

    @property
    def vertical_domain(self) -> List[str]:
        """List of unqiue object vertical domain"""
        return self._get_field_values("data.vertical_domain")

    @property
    async def vertical_domain_async(self) -> List[str]:
        """List of unqiue object vertical domain"""
        return await self._get_field_values_async("data.vertical_domain")

    @property
    def contents(self) -> List[str]:
        """List of unique contents"""
        return self._get_field_values("data.content.keyword")

    @property
    async def contents_async(self) -> List[str]:
        """List of unique contents"""
        return self._get_field_values_async("data.content.keyword")

    def _init_query(self, doc_type: str, query: Dict = None) -> Dict:
        new_query = super()._init_query(doc_type, query)
        case_filter = {
            "bool": {
                "must": [
                    {"term": {"_sumo.parent_object.keyword": self._case_uuid}}
                ]
            }
        }

        return self._utils.extend_query_object(new_query, case_filter)

    def _add_filter(
        self,
        name: Union[str, List[str], bool] = None,
        tagname: Union[str, List[str], bool] = None,
        dataformat: Union[str, List[str], bool] = None,
        iteration: Union[str, List[str], bool] = None,
        realization: Union[int, List[int], bool] = None,
        aggregation: Union[str, List[str], bool] = None,
        stage: Union[str, List[str], bool] = None,
        column: Union[str, List[str], bool] = None,
        time: TimeFilter = None,
        uuid: Union[str, List[str], bool] = None,
        stratigraphic: Union[str, List[str], bool] = None,
        vertical_domain: Union[str, List[str], bool] = None,
        content: Union[str, List[str], bool] = None,
        is_observation: bool = None,
        is_prediction: bool = None,
    ):
        must = []
        must_not = []

        prop_map = {
            "data.name.keyword": name,
            "data.tagname.keyword": tagname,
            "data.format": dataformat,
            "fmu.iteration.name.keyword": iteration,
            "fmu.realization.id": realization,
            "fmu.aggregation.operation.keyword": aggregation,
            "fmu.context.stage.keyword": stage,
            "data.spec.columns.keyword": column,
            "_id": uuid,
            "data.vertical_domain.keyword": vertical_domain,
            "data.content.keyword": content,
        }

        for prop, value in prop_map.items():
            if value is not None:
                if isinstance(value, bool):
                    if value:
                        must.append({"exists": {"field": prop}})
                    else:
                        must_not.append({"exists": {"field": prop}})
                else:
                    term = "terms" if isinstance(value, list) else "term"
                    must.append({term: {prop: value}})

        bool_prop_map = {
            "data.stratigraphic": stratigraphic,
            "data.is_observation": is_observation,
            "data.is_prediction": is_prediction,
        }
        for prop, value in bool_prop_map.items():
            if value is not None:
                must.append({"term": {prop: value}})

        query = {"bool": {}}

        if len(must) > 0:
            query["bool"]["must"] = must

        if len(must_not) > 0:
            query["bool"]["must_not"] = must_not

        if time:
            query = self._utils.extend_query_object(query, time._get_query())

        return super()._add_filter(query)




# Filter that matches 4d-seismic objects.

seismic4d = {
    "bool": {
        "must": [
            {
                "term": {
                    "data.content.keyword": "seismic"
                }
            },
            {
                "term": {
                    "data.time.t0.label.keyword": "base"
                }
            },
            {
                "term": {
                    "data.time.t1.label.keyword": "monitor"
                }
            }
        ]
    }
}



"""module containing class for child object"""

from typing import Dict
from io import BytesIO
from sumo.wrapper import SumoClient
from fmu.sumo.explorer.objects._document import Document


class Child(Document):
    """Class representing a child object in Sumo"""

    def __init__(self, sumo: SumoClient, metadata: Dict) -> None:
        """
        Args:
            sumo (SumoClient): connection to Sumo
            metadata: (dict): child object metadata
        """
        super().__init__(metadata)
        self._sumo = sumo
        self._blob = None

    @property
    def name(self) -> str:
        """Object name"""
        return self._get_property(["data", "name"])

    @property
    def content(self) -> str:
        """Content"""
        return self._get_property(["data", "content"])

    @property
    def tagname(self) -> str:
        """Object tagname"""
        return self._get_property(["data", "tagname"])

    @property
    def stratigraphic(self) -> str:
        """Object stratigraphic"""
        return self._get_property(["data", "stratigraphic"])

    @property
    def vertical_domain(self) -> str:
        """Object vertical_domain"""
        return self._get_property(["data", "vertical_domain"])

    @property
    def context(self) -> str:
        """Object context"""
        return self._get_property(["fmu", "context", "stage"])

    @property
    def iteration(self) -> int:
        """Object iteration"""
        return self._get_property(["fmu", "iteration", "name"])

    @property
    def realization(self) -> int:
        """Object realization"""
        return self._get_property(["fmu", "realization", "id"])

    @property
    def aggregation(self) -> str:
        """Object aggregation operation"""
        return self._get_property(["fmu", "aggregation", "operation"])

    @property
    def stage(self) -> str:
        """Object stage"""
        return self._get_property(["fmu", "context", "stage"])

    @property
    def format(self) -> str:
        """Object file format"""
        # (Legacy) alias for `dataformat`. Deprecate at some point?
        return self.dataformat

    @property
    def dataformat(self) -> str:
        """Object file format"""
        return self._get_property(["data", "format"])

    @property
    def relative_path(self) -> str:
        """Object relative file path"""
        return self._get_property(["file", "relative_path"])

    @property
    def blob(self) -> BytesIO:
        """Object blob"""
        if self._blob is None:
            res = self._sumo.get(f"/objects('{self.uuid}')/blob")
            self._blob = BytesIO(res.content)

        return self._blob

    @property
    async def blob_async(self) -> BytesIO:
        """Object blob"""
        if self._blob is None:
            res = await self._sumo.get_async(f"/objects('{self.uuid}')/blob")
            self._blob = BytesIO(res.content)

        return self._blob



"""Module containing class for collection of cubes """
from typing import Union, List, Dict, Tuple
from sumo.wrapper import SumoClient
from fmu.sumo.explorer.objects._child_collection import ChildCollection
from fmu.sumo.explorer.objects.cube import Cube
from fmu.sumo.explorer.pit import Pit
from fmu.sumo.explorer.timefilter import TimeFilter

TIMESTAMP_QUERY = {
    "bool": {
        "must": [{"exists": {"field": "data.time.t0"}}],
        "must_not": [{"exists": {"field": "data.time.t1"}}],
    }
}


class CubeCollection(ChildCollection):
    """Class for representing a collection of seismic cube objects in Sumo"""

    def __init__(
        self,
        sumo: SumoClient,
        case_uuid: str,
        query: Dict = None,
        pit: Pit = None,
    ):
        """
        Args:
            sumo (SumoClient): connection to Sumo
            case_uuid (str): parent case uuid
            query (dict): elastic query object
            pit (Pit): point in time
        """
        super().__init__("cube", sumo, case_uuid, query, pit)

    def __getitem__(self, index) -> Cube:
        doc = super().__getitem__(index)
        return Cube(self._sumo, doc)

    async def getitem_async(self, index: int) -> Cube:
        doc = await super().getitem_async(index)
        return Cube(self._sumo, doc)

    @property
    def timestamps(self) -> List[str]:
        """List of unique timestamps in CubeCollection"""
        return self._get_field_values(
            "data.time.t0.value", TIMESTAMP_QUERY, True
        )

    @property
    async def timestamps_async(self) -> List[str]:
        """List of unique timestamps in CubeCollection"""
        return await self._get_field_values_async(
            "data.time.t0.value", TIMESTAMP_QUERY, True
        )

    @property
    def intervals(self) -> List[Tuple]:
        """List of unique intervals in CubeCollection"""
        res = self._sumo.post(
            "/search",
            json={
                "query": self._query,
                "aggs": {
                    "t0": {
                        "terms": {"field": "data.time.t0.value", "size": 50},
                        "aggs": {
                            "t1": {
                                "terms": {
                                    "field": "data.time.t1.value",
                                    "size": 50,
                                }
                            }
                        },
                    }
                },
            },
        )

        buckets = res.json()["aggregations"]["t0"]["buckets"]
        intervals = []

        for bucket in buckets:
            t0 = bucket["key_as_string"]

            for t1 in bucket["t1"]["buckets"]:
                intervals.append((t0, t1["key_as_string"]))

        return intervals

    @property
    async def intervals_async(self) -> List[Tuple]:
        """List of unique intervals in CubeCollection"""
        res = await self._sumo.post_async(
            "/search",
            json={
                "query": self._query,
                "aggs": {
                    "t0": {
                        "terms": {"field": "data.time.t0.value", "size": 50},
                        "aggs": {
                            "t1": {
                                "terms": {
                                    "field": "data.time.t1.value",
                                    "size": 50,
                                }
                            }
                        },
                    }
                },
            },
        )

        buckets = res.json()["aggregations"]["t0"]["buckets"]
        intervals = []

        for bucket in buckets:
            t0 = bucket["key_as_string"]

            for t1 in bucket["t1"]["buckets"]:
                intervals.append((t0, t1["key_as_string"]))

        return intervals

    def filter(
        self,
        name: Union[str, List[str], bool] = None,
        tagname: Union[str, List[str], bool] = None,
        iteration: Union[str, List[str], bool] = None,
        realization: Union[int, List[int], bool] = None,
        stage: Union[str, List[str], bool] = None,
        time: TimeFilter = None,
        uuid: Union[str, List[str], bool] = None,
        is_observation: bool = None,
        is_prediction: bool = None,
        content: Union[str, List[str], bool] = None,
    ) -> "CubeCollection":
        """Filter cubes

        Args:
            name (Union[str, List[str], bool]): cube name
            tagname (Union[str, List[str], bool]): cube tagname
            iteration (Union[int, List[int], bool]): iteration id
            realization Union[int, List[int], bool]: realization id
            stage (Union[str, List[str], bool]): context/stage
            time (TimeFilter): time filter
            uuid (Union[str, List[str], bool]): cube object uuid
            is_observation (bool): cube is_observation
            is_prediction (bool): cube is_prediction
            content (Union[str, List[str], bool]): cube content

        Returns:
            CubeCollection: A filtered CubeCollection
        """
        query = super()._add_filter(
            name=name,
            tagname=tagname,
            iteration=iteration,
            realization=realization,
            stage=stage,
            time=time,
            uuid=uuid,
            is_observation=is_observation,
            is_prediction=is_prediction,
            content=content
        )

        return CubeCollection(self._sumo, self._case_uuid, query, self._pit)



"""Module containing utility class"""
from typing import List, Dict
import json
from sumo.wrapper import SumoClient

def _build_bucket_query(query, field):
    return {
        "size": 0,
        "query": query,
        "aggs": {
            f"{field}": {
                "composite": {
                    "size": 1000,
                    "sources": [
                        {
                            f"{field}": {
                                "terms": {
                                    "field": field
                                }
                            }
                        }
                    ]
                }
            }
        }
    }

def _set_after_key(query, field, after_key):
    if after_key is not None:
        query["aggs"][field]["composite"]["after"] = after_key
        pass
    return query

class Utils:
    """A class with utility functions for communicating with Sumo API"""

    def __init__(self, sumo: SumoClient) -> None:
        self._sumo = sumo

    def get_buckets(
        self,
        field: str,
        query: Dict,
        sort: List = None,
    ) -> List[Dict]:
        """Get a List of buckets

        Arguments:
            - field (str): a field in the metadata
            - query (List[Dict] or None): filter options
            - sort (List or None): sorting options

        Returns:
            A List of unique values for a given field
        """

        query = _build_bucket_query(query, field)
        all_buckets = []
        after_key = None
        while True:
            query = _set_after_key(query, field, after_key)
            res = self._sumo.post("/search", json=query).json()
            buckets = res["aggregations"][field]["buckets"]
            if len(buckets) == 0:
                break
            after_key = res["aggregations"][field]["after_key"]
            buckets = [{"key": bucket["key"][field], "doc_count": bucket["doc_count"]} for bucket in buckets]
            all_buckets = all_buckets + buckets
            pass

        return all_buckets

    async def get_buckets_async(
        self,
        field: str,
        query: Dict,
        sort: List = None,
    ) -> List[Dict]:
        """Get a List of buckets

        Arguments:
            - field (str): a field in the metadata
            - query (List[Dict] or None): filter options
            - sort (List or None): sorting options

        Returns:
            A List of unique values for a given field
        """

        query = _build_bucket_query(query, field)
        all_buckets = []
        after_key = None
        while True:
            query = _set_after_key(query, field, after_key)
            res = await self._sumo.post_async("/search", json=query)
            res = res.json()
            buckets = res["aggregations"][field]["buckets"]
            if len(buckets) == 0:
                break
            after_key = res["aggregations"][field]["after_key"]
            buckets = [{"key": bucket["key"][field], "doc_count": bucket["doc_count"]} for bucket in buckets]
            all_buckets = all_buckets + buckets
            pass

        return all_buckets

    def get_objects(
        self,
        size: int,
        query: Dict,
        select: List[str] = None,
    ) -> List[Dict]:
        """Get objects

        Args:
            size (int): number of objects to return
            query (List[Dict] or None): filter options
            select (List[str] or None): list of metadata fields to return

        Returns:
            List[Dict]: A List of metadata
        """
        query = {"size": size, "query": query}

        if select is not None:
            query["_source"] = select

        res = self._sumo.post("/search", json=query)

        return res.json()["hits"]["hits"]

    async def get_objects_async(
        self,
        size: int,
        query: Dict,
        select: List[str] = None,
    ) -> List[Dict]:
        """Get objects

        Args:
            size (int): number of objects to return
            query (List[Dict] or None): filter options
            select (List[str] or None): list of metadata fields to return

        Returns:
            List[Dict]: A List of metadata
        """
        query = {"size": size, "query": query}

        if select is not None:
            query["_source"] = select

        res = await self._sumo.post_async("/search", json=query)

        return res.json()["hits"]["hits"]

    def get_object(self, uuid: str, select: List[str] = None) -> Dict:
        """Get metadata object by uuid

        Args:
            uuid (str): uuid of metadata object
            select (List[str]): list of metadata fields to return

        Returns:
            Dict: a metadata object
        """

        query = {
            "query": {"term": {"_id": uuid}},
            "size": 1,
        }

        if select is not None:
            query["_source"] = select

        res = self._sumo.post("/search", json=query)
        hits = res.json()["hits"]["hits"]

        if len(hits) == 0:
            raise Exception(f"Document not found: {uuid}")

        return hits[0]

    async def get_object_async(self, uuid: str, select: List[str] = None) -> Dict:
        """Get metadata object by uuid

        Args:
            uuid (str): uuid of metadata object
            select (List[str]): list of metadata fields to return

        Returns:
            Dict: a metadata object
        """

        query = {
            "query": {"term": {"_id": uuid}},
            "size": 1,
        }

        if select is not None:
            query["_source"] = select

        res = await self._sumo.post_async("/search", json=query)
        hits = res.json()["hits"]["hits"]

        if len(hits) == 0:
            raise Exception(f"Document not found: {uuid}")

        return hits[0]

    def extend_query_object(self, old: Dict, new: Dict) -> Dict:
        """Extend query object

        Args:
            old (Dict): old query object
            new (Dict): new query object

        Returns:
            Dict: Extended query object
        """
        return_value = old
        if new is not None:
            stringified = json.dumps(old)
            extended = json.loads(stringified)

            for key in new:
                if key in extended:
                    if isinstance(new[key], dict):
                        extended[key] = self.extend_query_object(
                            extended[key], new[key]
                        )
                    elif isinstance(new[key], list):
                        for val in new[key]:
                            if val not in extended[key]:
                                extended[key].append(val)
                    else:
                        extended[key] = new[key]
                else:
                    extended[key] = new[key]
            return_value = extended
        return return_value

    def build_terms(self, keys_vals: Dict) -> List[Dict]:
        """Build a list of term objects

        `key_vals` expects a dictionary with the following structure:

        ```
        {
            "my.property.1": "value",
            "my.property.2": ["list", "of", "values"]
        }
        ```

        Returns a list of term expressions:

        ```
        [
            {"term": {"my.property.1": "value"}},
            {"terms": {"my.property.2": ["list", "of", "values"]}}
        ]
        ```

        Args:
            key_vals: Dict

        Returns:
            List[Dict]: list of term expressions
        """
        terms = []

        for key in keys_vals:
            val = keys_vals[key]
            if val is not None:
                items = [val] if not isinstance(val, list) else val
                terms.append({"terms": {key: items}})

        return terms



"""Module containing class for collection of documents"""
from typing import List, Dict
from sumo.wrapper import SumoClient
from fmu.sumo.explorer._utils import Utils
from fmu.sumo.explorer.pit import Pit


class DocumentCollection:
    """Class for representing a collection of documents in Sumo"""

    def __init__(
        self,
        doc_type: str,
        sumo: SumoClient,
        query: Dict = None,
        select: List[str] = None,
        pit: Pit = None,
    ):
        self._utils = Utils(sumo)
        self._type = doc_type
        self._sumo = sumo
        self._query = self._init_query(doc_type, query)

        self._pit = pit
        self._new_pit_id = None
        self._after = None
        self._curr_index = 0
        self._len = None
        self._items = []
        self._field_values = {}
        self._select = select

    def __iter__(self):
        self._curr_index = 0
        return self

    def __next__(self):
        if self._curr_index < self.__len__():
            res = self.__getitem__(self._curr_index)
            self._curr_index += 1
            return res
        else:
            raise StopIteration

    def __aiter__(self):
        self._curr_index = 0
        return self

    async def __anext__(self):
        if self._curr_index < await self.length_async():
            res = await self.getitem_async(self._curr_index)
            self._curr_index += 1
            return res
        else:
            raise StopAsyncIteration

    def __len__(self) -> int:
        """Get size of document collection

        Returns:
            Document collection size
        """
        if self._len is None:
            self._next_batch()

        return self._len

    async def length_async(self) -> int:
        """Get size of document collection. Async equivalent to 'len(collection)'

        Returns:
            Document collection size
        """
        if self._len is None:
            await self._next_batch_async()

        return self._len

    def __getitem__(self, index: int) -> Dict:
        """Get document

        Arguments:
            - index (int): index

        Returns:
            A document at a given index
        """
        if index > self.__len__():
            raise IndexError

        while len(self._items) <= index:
            hits_size = self._next_batch()

            if hits_size == 0:
                raise IndexError

        return self._items[index]

    async def getitem_async(self, index: int) -> Dict:
        """Get document. Async equivalent to 'collection[index]'

        Arguments:
            - index (int): index

        Returns:
            A document at a given index
        """
        if index > await self.length_async():
            raise IndexError

        while len(self._items) <= index:
            hits_size = await self._next_batch_async()

            if hits_size == 0:
                raise IndexError

        return self._items[index]

    def _get_field_values(
        self, field: str, query: Dict = None, key_as_string: bool = False
    ) -> List:
        """Get List of unique values for a given field

        Arguments:
            - field (str): a metadata field

        Returns:
            A List of unique values for the given field
        """
        if field not in self._field_values:
            bucket_query = self._utils.extend_query_object(self._query, query)
            key = "key_as_string" if key_as_string is True else "key"
            buckets = self._utils.get_buckets(field, bucket_query)
            self._field_values[field] = list(
                map(lambda bucket: bucket[key], buckets)
            )

        return self._field_values[field]

    async def _get_field_values_async(
        self, field: str, query: Dict = None, key_as_string: bool = False
    ) -> List:
        """Get List of unique values for a given field

        Arguments:
            - field (str): a metadata field

        Returns:
            A List of unique values for the given field
        """
        if field not in self._field_values:
            bucket_query = self._utils.extend_query_object(self._query, query)
            key = "key_as_string" if key_as_string is True else "key"
            buckets = await self._utils.get_buckets_async(field, bucket_query)
            self._field_values[field] = list(
                map(lambda bucket: bucket[key], buckets)
            )

        return self._field_values[field]

    def _next_batch(self) -> List[Dict]:
        """Get next batch of documents

        Returns:
            The next batch of documents
        """
        query = {
            "query": self._query,
            "sort": [{"_doc": {"order": "desc"}}],
            "size": 500,
        }

        if self._select:
            query["_source"] = self._select

        if self._len is None:
            query["track_total_hits"] = True

        if self._after is not None:
            query["search_after"] = self._after

        pit = None
        if self._pit is not None:
            pit = self._pit.get_pit_object(self._new_pit_id)
            query["pit"] = pit

        res = self._sumo.post("/search", json=query).json()
        hits = res["hits"]

        self._postprocess_batch(hits["hits"], pit)

        if self._pit is not None:
            self._new_pit_id = res["pit_id"]

        if self._len is None:
            self._len = hits["total"]["value"]

        if len(hits["hits"]) > 0:
            self._after = hits["hits"][-1]["sort"]
            self._items.extend(hits["hits"])

        return len(hits["hits"])

    async def _next_batch_async(self) -> List[Dict]:
        """Get next batch of documents

        Returns:
            The next batch of documents
        """
        query = {
            "query": self._query,
            "sort": [{"_doc": {"order": "desc"}}],
            "size": 500,
        }

        if self._select:
            query["_source"] = self._select

        if self._len is None:
            query["track_total_hits"] = True

        if self._after is not None:
            query["search_after"] = self._after

        pit = None
        if self._pit is not None:
            pit = self._pit.get_pit_object(self._new_pit_id)
            query["pit"] = pit

        res = await self._sumo.post_async("/search", json=query)
        data = res.json()
        hits = data["hits"]

        await self._postprocess_batch_async(hits["hits"], pit)

        if self._pit is not None:
            self._new_pit_id = data["pit_id"]

        if self._len is None:
            self._len = hits["total"]["value"]

        if len(hits["hits"]) > 0:
            self._after = hits["hits"][-1]["sort"]
            self._items.extend(hits["hits"])

        return len(hits["hits"])

    def _postprocess_batch(self, hits, pit):
        return

    async def _postprocess_batch_async(self, hits, pit):
        return

    def _init_query(self, doc_type: str, query: Dict = None) -> Dict:
        """Initialize base filter for document collection

        Arguments:
            - type (str): object type
            - filters (List[Dict]): a List of filters

        Returns:
            Document collection base filters
        """
        class_filter = {
            "bool": {"must": [{"term": {"class.keyword": doc_type}}]}
        }

        if query is not None:
            return self._utils.extend_query_object(class_filter, query)

        return class_filter

    def _add_filter(self, query: Dict) -> Dict:
        """Add filter to DocumentCollection base filter

        Argmuments:
            - user_filter (Dict[str, List]): new filters

        Returns:
            Filter object containing base filters and new filters
        """

        return self._utils.extend_query_object(self._query, query)



"""Test table objects.

  * Table
  * AggregatedTable
  * TableCollection

"""
import pandas as pd
import pyarrow as pa
from fmu.sumo.explorer import Explorer, AggregatedTable
import pytest

# Fixed test case ("Drogon_AHM_2023-02-22") in Sumo/DEV
TESTCASE_UUID = "10f41041-2c17-4374-a735-bb0de62e29dc"

@pytest.fixture(name="explorer")
def fixture_explorer(token: str) -> Explorer:
    """Returns explorer"""
    return Explorer("dev", token=token)

@pytest.fixture(name="case")
def fixture_case(explorer: Explorer):
    """Return fixed testcase."""
    return explorer.get_case_by_uuid(TESTCASE_UUID)

@pytest.fixture(name="table")
def fixture_table(case):
    """Get one table for further testing."""
    return case.tables[0]
    
### Table

def test_table_to_pandas(table):
    """Test the to_pandas method."""
    df = table.to_pandas()
    assert isinstance(df, pd.DataFrame)

def test_table_to_arrow(table):
    """Test the to_arrow() method"""
    arrow = table.to_arrow()
    assert isinstance(arrow, pa.Table)


### Aggregated Table

def test_aggregated_summary_arrow(case):
    """Test usage of Aggregated class with default type"""

    table = AggregatedTable(case, "summary", "eclipse", "iter-0")

    assert len(table.columns) == 972 + 2
    column = table["FOPT"]

    assert isinstance(column.to_arrow(), pa.Table)
    with pytest.raises(IndexError) as e_info:
        table = table["banana"]
        assert (
            e_info.value.args[0] == "Column: 'banana' does not exist try again"
        )


def test_aggregated_summary_pandas(case):
    """Test usage of Aggregated class with item_type=pandas"""
    table = AggregatedTable(case, "summary", "eclipse", "iter-0")
    assert isinstance(table["FOPT"].to_pandas(), pd.DataFrame)


def test_get_fmu_iteration_parameters(case):
    """Test getting the metadata of of an object"""
    table = AggregatedTable(case, "summary", "eclipse", "iter-0")
    assert isinstance(table.parameters, dict)



"""Sumo cases and child objects"""
from fmu.sumo.explorer.objects.case import Case
from fmu.sumo.explorer.objects.case_collection import CaseCollection
from fmu.sumo.explorer.objects.surface import Surface
from fmu.sumo.explorer.objects.surface_collection import SurfaceCollection
from fmu.sumo.explorer.objects.polygons import Polygons
from fmu.sumo.explorer.objects.polygons_collection import PolygonsCollection
from fmu.sumo.explorer.objects.table import Table
from fmu.sumo.explorer.objects.table_collection import TableCollection



"""Module containing class for collection of tables"""
from typing import Union, List, Dict
from sumo.wrapper import SumoClient
from fmu.sumo.explorer.objects._child_collection import ChildCollection
from fmu.sumo.explorer.objects.table import Table
from fmu.sumo.explorer.pit import Pit


class TableCollection(ChildCollection):
    """Class for representing a collection of table objects in Sumo"""

    def __init__(
        self,
        sumo: SumoClient,
        case_uuid: str,
        query: Dict = None,
        pit: Pit = None,
    ):
        """
        Args:
            sumo (SumoClient): connection to Sumo
            case_uuid (str): parent case uuid
            query (dict): elastic query object
            pit (Pit): point in time
        """
        super().__init__("table", sumo, case_uuid, query, pit)

    def __getitem__(self, index) -> Table:
        doc = super().__getitem__(index)
        return Table(self._sumo, doc)

    async def getitem_async(self, index: int) -> Table:
        doc = await super().getitem_async(index)
        return Table(self._sumo, doc)

    @property
    def columns(self) -> List[str]:
        """List of unique column names"""
        return self._get_field_values("data.spec.columns.keyword")

    @property
    async def columns_async(self) -> List[str]:
        """List of unique column names"""
        return await self._get_field_values_async("data.spec.columns.keyword")

    def filter(
        self,
        name: Union[str, List[str], bool] = None,
        tagname: Union[str, List[str], bool] = None,
        iteration: Union[str, List[str], bool] = None,
        realization: Union[int, List[int], bool] = None,
        aggregation: Union[str, List[str], bool] = None,
        stage: Union[str, List[str], bool] = None,
        column: Union[str, List[str], bool] = None,
        uuid: Union[str, List[str], bool] = None,
        content: Union[str, List[str], bool] = None,
    ) -> "TableCollection":
        """Filter tables

        Arguments:
            name (Union[str, List[str], bool]): table name
            tagname (Union[str, List[str], bool]): table tagname
            iteration (Union[int, List[int], bool]): iteration id
            realization Union[int, List[int], bool]: realization id
            aggregation (Union[str, List[str], bool]): aggregation operation
            stage (Union[str, List[str], bool]): context/stage
            uuid (Union[str, List[str], bool]): table object uuid
            content (Union[str, List[str], bool): table content

        Returns:
            TableCollection: A filtered TableCollection
        """

        query = super()._add_filter(
            name=name,
            tagname=tagname,
            iteration=iteration,
            realization=realization,
            aggregation=aggregation,
            stage=stage,
            column=column,
            uuid=uuid,
            content=content
        )
        return TableCollection(self._sumo, self._case_uuid, query, self._pit)



"""Module containg class for surface"""
from typing import Dict
from xtgeo import RegularSurface, surface_from_file
from fmu.sumo.explorer.objects._child import Child


class Surface(Child):
    """Class representing a surface object in Sumo"""

    @property
    def bbox(self) -> Dict:
        """Surface bbox data"""
        return self._get_property(["data", "bbox"])

    @property
    def spec(self) -> Dict:
        """Surface spec data"""
        return self._get_property(["data", "spec"])

    @property
    def timestamp(self) -> str:
        """Surface timestmap data"""
        t0 = self._get_property(["data", "time", "t0", "value"])
        t1 = self._get_property(["data", "time", "t1", "value"])

        if t0 is not None and t1 is None:
            return t0

        return None

    @property
    def interval(self) -> str:
        """Surface interval data"""
        t0 = self._get_property(["data", "time", "t0", "value"])
        t1 = self._get_property(["data", "time", "t1", "value"])

        if t0 is not None and t1 is not None:
            return (t0, t1)

        return None

    def to_regular_surface(self) -> RegularSurface:
        """Get surface object as a RegularSurface

        Returns:
            RegularSurface: A RegularSurface object
        """
        try:
            return surface_from_file(self.blob)
        except TypeError as type_err:
            raise TypeError(f"Unknown format: {self.format}") from type_err

    async def to_regular_surface_async(self) -> RegularSurface:
        """Get surface object as a RegularSurface

        Returns:
            RegularSurface: A RegularSurface object
        """
        try:
            return surface_from_file(await self.blob_async)
        except TypeError as type_err:
            raise TypeError(f"Unknown format: {self.format}") from type_err



"""Module containing class for collection of surfaces"""

from typing import Union, List, Dict, Tuple
from io import BytesIO
from xtgeo import RegularSurface, surface_from_file
from sumo.wrapper import SumoClient
from fmu.sumo.explorer.objects._child_collection import ChildCollection
from fmu.sumo.explorer.objects.surface import Surface
from fmu.sumo.explorer.timefilter import TimeFilter
from fmu.sumo.explorer.pit import Pit

TIMESTAMP_QUERY = {
    "bool": {
        "must": [{"exists": {"field": "data.time.t0"}}],
        "must_not": [{"exists": {"field": "data.time.t1"}}],
    }
}


class SurfaceCollection(ChildCollection):
    """Class representing a collection of surface objects in Sumo"""

    def __init__(
        self,
        sumo: SumoClient,
        case_uuid: str,
        query: Dict = None,
        pit: Pit = None,
    ):
        """
        Args:
            sumo (SumoClient): connection to Sumo
            case_uuid (str): parent case uuid
            query (dict): elastic query object
            pit (Pit): point in time
        """
        super().__init__("surface", sumo, case_uuid, query, pit)

        self._aggregation_cache = {}

    def __getitem__(self, index) -> Surface:
        doc = super().__getitem__(index)
        return Surface(self._sumo, doc)

    async def getitem_async(self, index: int) -> Surface:
        doc = await super().getitem_async(index)
        return Surface(self._sumo, doc)

    @property
    def timestamps(self) -> List[str]:
        """List of unique timestamps in SurfaceCollection"""
        return self._get_field_values(
            "data.time.t0.value", TIMESTAMP_QUERY, True
        )

    @property
    async def timestamps_async(self) -> List[str]:
        """List of unique timestamps in CubeCollection"""
        return await self._get_field_values_async(
            "data.time.t0.value", TIMESTAMP_QUERY, True
        )

    @property
    def intervals(self) -> List[Tuple]:
        """List of unique intervals in SurfaceCollection"""
        res = self._sumo.post(
            "/search",
            json={
                "query": self._query,
                "aggs": {
                    "t0": {
                        "terms": {"field": "data.time.t0.value", "size": 50},
                        "aggs": {
                            "t1": {
                                "terms": {
                                    "field": "data.time.t1.value",
                                    "size": 50,
                                }
                            }
                        },
                    }
                },
            },
        )

        buckets = res.json()["aggregations"]["t0"]["buckets"]
        intervals = []

        for bucket in buckets:
            t0 = bucket["key_as_string"]

            for t1 in bucket["t1"]["buckets"]:
                intervals.append((t0, t1["key_as_string"]))

        return intervals

    @property
    async def intervals_async(self) -> List[Tuple]:
        """List of unique intervals in SurfaceCollection"""
        res = await self._sumo.post_async(
            "/search",
            json={
                "query": self._query,
                "aggs": {
                    "t0": {
                        "terms": {"field": "data.time.t0.value", "size": 50},
                        "aggs": {
                            "t1": {
                                "terms": {
                                    "field": "data.time.t1.value",
                                    "size": 50,
                                }
                            }
                        },
                    }
                },
            },
        )

        buckets = res.json()["aggregations"]["t0"]["buckets"]
        intervals = []

        for bucket in buckets:
            t0 = bucket["key_as_string"]

            for t1 in bucket["t1"]["buckets"]:
                intervals.append((t0, t1["key_as_string"]))

        return intervals

    def _aggregate(self, operation: str) -> RegularSurface:
        if operation not in self._aggregation_cache:
            objects = self._utils.get_objects(500, self._query, ["_id"])
            object_ids = list(map(lambda obj: obj["_id"], objects))

            res = self._sumo.post(
                "/aggregations",
                json={
                    "operation": [operation],
                    "object_ids": object_ids,
                    "class": "surface",
                },
            )

            self._aggregation_cache[operation] = surface_from_file(
                BytesIO(res.content)
            )

        return self._aggregation_cache[operation]

    async def _aggregate_async(self, operation: str) -> RegularSurface:
        if operation not in self._aggregation_cache:
            objects = await self._utils.get_objects_async(
                500, self._query, ["_id"]
            )
            object_ids = list(map(lambda obj: obj["_id"], objects))

            res = await self._sumo.post_async(
                "/aggregations",
                json={
                    "operations": [operation],
                    "object_ids": object_ids,
                    "class": "surface",
                },
            )

            self._aggregation_cache[operation] = surface_from_file(
                BytesIO(res.content)
            )

        return self._aggregation_cache[operation]

    def filter(
        self,
        name: Union[str, List[str], bool] = None,
        tagname: Union[str, List[str], bool] = None,
        dataformat: Union[str, List[str], bool] = None,
        stratigraphic: Union[str, List[str], bool] = None,
        vertical_domain: Union[str, List[str], bool] = None,
        iteration: Union[str, List[str], bool] = None,
        realization: Union[int, List[int], bool] = None,
        aggregation: Union[str, List[str], bool] = None,
        stage: Union[str, List[str], bool] = None,
        time: TimeFilter = None,
        uuid: Union[str, List[str], bool] = None,
        content: Union[str, List[str], bool] = None,
        is_observation: bool = None,
        is_prediction: bool = None,
    ) -> "SurfaceCollection":
        """Filter surfaces

        Apply filters to the SurfaceCollection and get a new filtered instance.

        Args:
            name (Union[str, List[str], bool]): surface name
            tagname (Union[str, List[str], bool]): surface tagname
            dataformat (Union[str, List[str], bool]): surface data format
            iteration (Union[int, List[int], bool]): iteration id
            realization Union[int, List[int], bool]: realization id
            aggregation (Union[str, List[str], bool]): aggregation operation
            stage (Union[str, List[str], bool]): context/stage
            time (TimeFilter): time filter
            uuid (Union[str, List[str], bool]): surface object uuid
            stratigraphic (Union[str, List[str], bool]): surface stratigraphic
            vertical_domain (Union[str, List[str], bool]): surface vertical_domain
            content (Union[str, List[str], bool): = surface content

        Returns:
            SurfaceCollection: A filtered SurfaceCollection

        Examples:

            Match one value::

                surfs = case.surfaces.filter(
                    iteration="iter-0"
                    name="my_surface_name"
                )

            Match multiple values::

                surfs = case.surfaces.filter(
                    name=["one_name", "another_name"]
                )

            Get aggregated surfaces with specific operation::

                surfs = case.surfaces.filter(
                    aggregation="max"
                )

            Get all aggregated surfaces::

                surfs = case.surfaces.filter(
                    aggregation=True
                )

            Get all non-aggregated surfaces::

                surfs = case.surfaces.filter(
                    aggregation=False
                )

        """

        query = super()._add_filter(
            name=name,
            tagname=tagname,
            dataformat=dataformat,
            iteration=iteration,
            realization=realization,
            aggregation=aggregation,
            stage=stage,
            time=time,
            uuid=uuid,
            stratigraphic=stratigraphic,
            vertical_domain=vertical_domain,
            content=content,
            is_observation=is_observation,
            is_prediction=is_prediction,
        )

        return SurfaceCollection(self._sumo, self._case_uuid, query, self._pit)

    def mean(self) -> RegularSurface:
        """Perform a mean aggregation"""
        return self._aggregate("mean")

    async def mean_async(self) -> RegularSurface:
        """Perform a mean aggregation"""
        return await self._aggregate_async("mean")

    def min(self) -> RegularSurface:
        """Perform a minimum aggregation"""
        return self._aggregate("min")

    async def min_async(self) -> RegularSurface:
        """Perform a minimum aggregation"""
        return await self._aggregate_async("min")

    def max(self) -> RegularSurface:
        """Perform a maximum aggregation"""
        return self._aggregate("max")

    async def max_async(self) -> RegularSurface:
        """Perform a maximum aggregation"""
        return await self._aggregate_async("max")

    def std(self) -> RegularSurface:
        """Perform a standard deviation aggregation"""
        return self._aggregate("std")

    async def std_async(self) -> RegularSurface:
        """Perform a standard deviation aggregation"""
        return await self._aggregate_async("std")

    def p10(self) -> RegularSurface:
        """Perform a percentile aggregation"""
        return self._aggregate("p10")

    async def p10_async(self) -> RegularSurface:
        """Perform a percentile aggregation"""
        return await self._aggregate_async("p10")

    def p50(self) -> RegularSurface:
        """Perform a percentile aggregation"""
        return self._aggregate("p50")

    async def p50_async(self) -> RegularSurface:
        """Perform a percentile aggregation"""
        return await self._aggregate_async("p50")

    def p90(self) -> RegularSurface:
        """Perform a percentile aggregation"""
        return self._aggregate("p90")

    async def p90_async(self) -> RegularSurface:
        """Perform a percentile aggregation"""
        return await self._aggregate_async("p90")



"""Module containing case class"""
from typing import Dict, List
from sumo.wrapper import SumoClient
from fmu.sumo.explorer.objects._document import Document
from fmu.sumo.explorer.objects.surface_collection import SurfaceCollection
from fmu.sumo.explorer.objects.polygons_collection import PolygonsCollection
from fmu.sumo.explorer.objects.table_collection import TableCollection
from fmu.sumo.explorer.objects.cube_collection import CubeCollection
from fmu.sumo.explorer.objects.dictionary_collection import (
    DictionaryCollection,
)
from fmu.sumo.explorer._utils import Utils
from fmu.sumo.explorer.pit import Pit


class Case(Document):
    """Class for representing a case in Sumo"""

    def __init__(self, sumo: SumoClient, metadata: Dict, overview: Dict,
                 pit: Pit = None):
        super().__init__(metadata)
        self._overview = overview
        self._pit = pit
        self._sumo = sumo
        self._utils = Utils(sumo)
        self._iterations = None

    @property
    def name(self) -> str:
        """Case name"""
        return self._get_property(["fmu", "case", "name"])

    @property
    def overview(self):
        """Overview of case contents."""
        return self._overview

    @property
    def status(self) -> str:
        """Case status"""
        return self._get_property(["_sumo", "status"])

    @property
    def user(self) -> str:
        """Name of user who uploaded the case"""
        return self._get_property(["fmu", "case", "user", "id"])

    @property
    def asset(self) -> str:
        """Case asset"""
        return self._get_property(["access", "asset", "name"])

    @property
    def field(self) -> str:
        """Case field"""
        fields = self._get_property(["masterdata", "smda", "field"])
        return fields[0]["identifier"]

    @property
    def iterations(self) -> List[Dict]:
        """List of case iterations"""
        if self._iterations is None:
            query = {
                "query": {"term": {"_sumo.parent_object.keyword": self.uuid}},
                "aggs": {
                    "uuid": {
                        "terms": {
                            "field": "fmu.iteration.uuid.keyword",
                            "size": 50,
                        },
                        "aggs": {
                            "name": {
                                "terms": {
                                    "field": "fmu.iteration.name.keyword",
                                    "size": 1,
                                }
                            },
                            "realizations": {
                                "cardinality": {
                                    "field": "fmu.realization.id",
                                }
                            },
                        },
                    },
                },
                "size": 0,
            }

            res = self._sumo.post("/search", json=query)
            buckets = res.json()["aggregations"]["uuid"]["buckets"]
            iterations = []

            for bucket in buckets:
                iterations.append(
                    {
                        "uuid": bucket["key"],
                        "name": bucket["name"]["buckets"][0]["key"],
                        "realizations": bucket["realizations"]["value"],
                    }
                )

            self._iterations = iterations

        return self._iterations

    @property
    async def iterations_async(self) -> List[Dict]:
        """List of case iterations"""
        if self._iterations is None:
            query = {
                "query": {"term": {"_sumo.parent_object.keyword": self.uuid}},
                "aggs": {
                    "id": {
                        "terms": {"field": "fmu.iteration.id", "size": 50},
                        "aggs": {
                            "name": {
                                "terms": {
                                    "field": "fmu.iteration.name.keyword",
                                    "size": 1,
                                }
                            },
                            "realizations": {
                                "terms": {
                                    "field": "fmu.realization.id",
                                    "size": 1000,
                                }
                            },
                        },
                    },
                },
                "size": 0,
            }

            res = await self._sumo.post_async("/search", json=query)
            buckets = res.json()["aggregations"]["id"]["buckets"]
            iterations = []

            for bucket in buckets:
                iterations.append(
                    {
                        "id": bucket["key"],
                        "name": bucket["name"]["buckets"][0]["key"],
                        "realizations": len(bucket["realizations"]["buckets"]),
                    }
                )

            self._iterations = iterations

        return self._iterations

    def get_realizations(self, iteration: str = None) -> List[int]:
        """Get a list of realization ids

        Calling this method without the iteration argument will
        return a list of unique realization ids across iterations.
        It is not guaranteed that all realizations in this list exists
        in all case iterations.

        Args:
            iteration (str): iteration name

        Returns:
            List[int]: realization ids
        """
        must = [{"term": {"_sumo.parent_object.keyword": self.uuid}}]

        if iteration:
            must.append({"term": {"fmu.iteration.name.keyword": iteration}})

        buckets = self._utils.get_buckets(
            "fmu.realization.id",
            query={"bool": {"must": must}},
            sort=["fmu.realization.id"],
        )

        return list(map(lambda b: b["key"], buckets))

    async def get_realizations_async(self, iteration: str = None) -> List[int]:
        """Get a list of realization ids

        Calling this method without the iteration argument will
        return a list of unique realization ids across iterations.
        It is not guaranteed that all realizations in this list exists
        in all case iterations.

        Args:
            iteration (str): iteration name

        Returns:
            List[int]: realization ids
        """
        must = [{"term": {"_sumo.parent_object.keyword": self.uuid}}]

        if iteration:
            must.append({"term": {"fmu.iteration.name.keyword": iteration}})

        buckets = await self._utils.get_buckets_async(
            "fmu.realization.id",
            query={"bool": {"must": must}},
            sort=["fmu.realization.id"],
        )

        return list(map(lambda b: b["key"], buckets))

    @property
    def surfaces(self) -> SurfaceCollection:
        """List of case surfaces"""
        return SurfaceCollection(self._sumo, self._uuid, pit=self._pit)

    @property
    def polygons(self) -> PolygonsCollection:
        """List of case polygons"""
        return PolygonsCollection(self._sumo, self._uuid, pit=self._pit)

    @property
    def tables(self) -> TableCollection:
        """List of case tables"""
        return TableCollection(self._sumo, self._uuid, pit=self._pit)

    @property
    def cubes(self) -> CubeCollection:
        """List of case tables"""
        return CubeCollection(self._sumo, self._uuid, pit=self._pit)

    @property
    def dictionaries(self) -> DictionaryCollection:
        """List of case dictionaries"""
        return DictionaryCollection(self._sumo, self._uuid, pit=self._pit)



"""Module containing class for collection of cases"""
from typing import Union, List, Dict
from sumo.wrapper import SumoClient
from fmu.sumo.explorer.objects._document_collection import DocumentCollection
from fmu.sumo.explorer.objects.case import Case
from fmu.sumo.explorer.pit import Pit

_CASE_FIELDS = {
    "include": [],
    "exclude": []
}

def _make_overview_query(ids, pit):
    query = {
        "query": {
            "terms": {
                "fmu.case.uuid.keyword": ids
            }
        },
        "aggs": {
            "cases": {
                "terms": {
                    "field": "fmu.case.uuid.keyword",
                    "size": 1000
                },
                "aggs": {
                    "iteration_uuids": {
                        "terms": {
                            "field": "fmu.iteration.uuid.keyword",
                            "size": 100
                        }
                    },
                    "iteration_names": {
                        "terms": {
                            "field": "fmu.iteration.name.keyword",
                            "size": 100
                        }
                    },
                    "data_types": {
                        "terms": {
                            "field": "class.keyword",
                            "size": 100
                        }
                    },
                    "iterations": {
                        "terms": {
                            "field": "fmu.iteration.uuid.keyword",
                            "size": 100
                        },
                        "aggs": {
                            "iteration_name": {
                                "terms": {
                                    "field": "fmu.iteration.name.keyword",
                                    "size": 100
                                }
                            },
                            "numreal": {
                                "cardinality": {
                                    "field": "fmu.realization.id"
                                }
                            },
                            "maxreal": {
                                "max": {
                                    "field": "fmu.realization.id"
                                }
                            },
                            "minreal": {
                                "min": {
                                    "field": "fmu.realization.id"
                                }
                            }
                        }
                    }
                }
            }
        },
        "size": 0
    }
    if pit:
        query["pit"] = pit
    return query


class CaseCollection(DocumentCollection):
    """A class for representing a collection of cases in Sumo"""

    def __init__(self, sumo: SumoClient, query: Dict = None, pit: Pit = None, has = None):
        """
        Args:
            sumo (SumoClient): connection to Sumo
            query (dict): elastic query object
            pit (Pit): point in time
            has (dict): query for specific child objects
        """
        super().__init__("case", sumo, query, _CASE_FIELDS, pit)
        self._overviews = {}
        self._has = has

    @property
    def names(self) -> List[str]:
        """List of unique case names"""
        return self._get_field_values("fmu.case.name.keyword")

    @property
    async def names_async(self) -> List[str]:
        """List of unique case names"""
        return await self._get_field_values_async("fmu.case.name.keyword")

    @property
    def uuids(self) -> List[str]:
        """List of unique case uuids"""
        return self._get_field_values("fmu.case.uuid.keyword")

    @property
    async def uuids_async(self) -> List[str]:
        """List of unique case uuids"""
        return await self._get_field_values_async("fmu.case.uuid.keyword")

    @property
    def statuses(self) -> List[str]:
        """List of unique statuses"""
        return self._get_field_values("_sumo.status.keyword")

    @property
    async def statuses_async(self) -> List[str]:
        """List of unique statuses"""
        return await self._get_field_values_async("_sumo.status.keyword")

    @property
    def users(self) -> List[str]:
        """List of unique user names"""
        return self._get_field_values("fmu.case.user.id.keyword")

    @property
    async def users_async(self) -> List[str]:
        """List of unique user names"""
        return await self._get_field_values_async("fmu.case.user.id.keyword")

    @property
    def assets(self) -> List[str]:
        """List of unique asset names"""
        return self._get_field_values("access.asset.name.keyword")

    @property
    async def assets_async(self) -> List[str]:
        """List of unique asset names"""
        return await self._get_field_values_async("access.asset.name.keyword")

    @property
    def fields(self) -> List[str]:
        """List of unique field names"""
        return self._get_field_values(
            "masterdata.smda.field.identifier.keyword"
        )

    @property
    async def fields_async(self) -> List[str]:
        """List of unique field names"""
        return await self._get_field_values_async(
            "masterdata.smda.field.identifier.keyword"
        )

    def __getitem__(self, index: int) -> Case:
        doc = super().__getitem__(index)
        uuid = doc["_id"]
        overview = self._overviews[uuid]
        return Case(self._sumo, doc, overview, self._pit)

    async def getitem_async(self, index: int) -> Case:
        doc = await super().getitem_async(index)
        uuid = doc["_id"]
        overview = self._overviews[uuid]
        return Case(self._sumo, doc, overview, self._pit)

    def _next_batch(self) -> List[Dict]:
        """Get next batch of documents

        Returns:
            The next batch of documents
        """
        if self._has is not None:
            uuids = self.uuids
            query = { "bool": { "must": [ {"terms": { "fmu.case.uuid.keyword": uuids}}, self._has]}}
            nuuids = [ x["key"] for x in self._utils.get_buckets("fmu.case.uuid.keyword", query)]
            self._query = {"ids": {"values": nuuids}}
            self._has = None
        return super()._next_batch()

    async def _next_batch_async(self) -> List[Dict]:
        """Get next batch of documents

        Returns:
            The next batch of documents
        """
        if self._has is not None:
            uuids = await self.uuids_async
            query = { "bool": { "must": [ {"terms": { "fmu.case.uuid.keyword": uuids}}, self._has]}}
            nuuids = [ x["key"] for x in self._utils.get_buckets("fmu.case.uuid.keyword", query)]
            self._query = {"ids": {"values": nuuids}}
            self._has = None
        return await super()._next_batch_async()

    def _postprocess_batch(self, hits, pit):
        ids = [hit["_id"] for hit in hits]
        query = _make_overview_query(ids, pit)
        res = self._sumo.post("/search", json=query)
        data = res.json()
        aggs = data["aggregations"]
        self._insert_overviews(aggs)
        return

    async def _postprocess_batch_async(self, hits, pit):
        ids = [hit["_id"] for hit in hits]
        query = _make_overview_query(ids, pit)
        res = await self._sumo.post_async("/search", json=query)
        data = res.json()
        aggs = data["aggregations"]
        self._insert_overviews(aggs)
        return

    def _insert_overviews(self, aggs):
        def extract_bucket_keys(bucket, name):
            return [b["key"] for b in bucket[name]["buckets"]]
        for bucket in aggs["cases"]["buckets"]:
            caseid = bucket["key"]
            iteration_names = extract_bucket_keys(bucket, "iteration_names")
            iteration_uuids = extract_bucket_keys(bucket, "iteration_uuids")
            data_types = extract_bucket_keys(bucket, "data_types")
            iterations = {}
            for ibucket in bucket["iterations"]["buckets"]:
                iterid = ibucket["key"]
                itername = extract_bucket_keys(ibucket, "iteration_name")
                minreal = ibucket["minreal"]["value"]
                maxreal = ibucket["maxreal"]["value"]
                numreal = ibucket["numreal"]["value"]
                iterations[iterid] = {
                    "name": itername,
                    "minreal": minreal,
                    "maxreal": maxreal,
                    "numreal": numreal
                }
            self._overviews[caseid] = {
                "iteration_names": iteration_names,
                "iteration_uuids": iteration_uuids,
                "data_types": data_types,
                "iterations": iterations
            }
        return

    def filter(
        self,
        uuid: Union[str, List[str]] = None,
        name: Union[str, List[str]] = None,
        status: Union[str, List[str]] = None,
        user: Union[int, List[int]] = None,
        asset: Union[int, List[int]] = None,
        field: Union[str, List[str]] = None,
        has: Dict = None,
    ) -> "CaseCollection":
        """Filter cases

        Args:
            uuid (str or List[str]): case uuid
            name (str or List[str]): case name
            status (str or List[str]): case status
            user (str or List[str]): name of case owner
            asset (str or List[str]): asset
            field (str or List[str]): field

        Returns:
            CaseCollection: A filtered CaseCollection
        """
        must = self._utils.build_terms(
            {
                "_id": uuid,
                "fmu.case.name.keyword": name,
                "_sumo.status.keyword": status,
                "fmu.case.user.id.keyword": user,
                "access.asset.name.keyword": asset,
                "masterdata.smda.field.identifier.keyword": field,
            }
        )

        query = super()._add_filter({"bool": {"must": must}})

        return CaseCollection(self._sumo, query, self._pit, has = has)
